/*******************************************************************************
* Copyright (C) 2019 Intel Corporation
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

/*
!  Content:
!      oneAPI Math Kernel Library (oneMKL) SYCL interface
!******************************************************************************/

#ifndef _MKL_SYCL_HPP_
#define _MKL_SYCL_HPP_

#include "mkl.h"
#include "oneapi/mkl/export.hpp"
#include "oneapi/mkl/types.hpp"
#include "oneapi/mkl/spec.hpp"
#include "oneapi/mkl/blas.hpp"
#include "oneapi/mkl/dft.hpp"
#include "oneapi/mkl/lapack.hpp"
#include "oneapi/mkl/vm.hpp"
#include "oneapi/mkl/rng.hpp"
#include "oneapi/mkl/spblas.hpp"
#include "oneapi/mkl/stats.hpp"

#endif /* _MKL_SYCL_HPP_ */
