/* file: oneapi/mkl/vm.hpp */
/*******************************************************************************
* Copyright (C) 2019 Intel Corporation
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

#ifndef _ONEAPI_MKL_VM_HPP_
#define _ONEAPI_MKL_VM_HPP_ 1

#include <cstdint>
#include <complex>
#include <exception>
#include <sycl/sycl.hpp>

#include "oneapi/mkl/vm/decls.hpp"
#include "oneapi/mkl/vm/buffer.hpp"
#include "oneapi/mkl/vm/usm.hpp"
#include "oneapi/mkl/vm/span.hpp"
#include "oneapi/mkl/vm/spec.hpp"


#endif // ifndef _ONEAPI_MKL_VM_HPP_
